/* FLENG - header-file for C part of runtime */

#ifndef FLENG_H
#define FLENG_H

#include <stdio.h>
#include <pthread.h>

#ifdef __linux__
#include <poll.h>
#endif

#define FL_PORT_SIZE    0x4000
#define FL_GLOBAL_TABLE 256

#define FL_TAG_MASK     0xff000000
#define FL_COUNT_MASK   0x00ffffff
#define FL_INT_BIT      1
#define FL_ATOM_BIT     2
#define FL_BITS_MASK    3

#define FL_LIST_TAG     0x01000000
#define FL_FLOAT_TAG    0x02000000
#define FL_VAR_TAG      0x04000000
#define FL_PORT_TAG     0x05000000
#define FL_REF_TAG      0x07000000
#define FL_MODULE_TAG   0x08000000
#define FL_INTERNED     0x7f000000

#define FL_TUPLE_BIT    0x80000000

#define FL_BROKEN_HEART 0xadde0000

typedef void *FL_VAL;

#define FL_FWD      -1
#define FL_BWD      -2
#define FL_NORTH    -3
#define FL_EAST     -4
#define FL_SOUTH    -5
#define FL_WEST     -6 
#define FL_ALL      -7 

typedef struct {
    unsigned long tag;
    FL_VAL car, cdr;
} FL_CELL;

typedef struct {
    void *addr;
    FL_VAL args[4];
    char *info;
} FL_GOAL;

typedef struct fl_htnode {
    FL_VAL key, val;
    struct fl_htnode *next;
} FL_HTNODE;

typedef struct {
    FL_GOAL *goal;
    FL_VAL *sstackp;
    long tailcalls;
    long timeslice;
    long ticks;
    void *stackbase;
    FL_CELL *freelist;
    FL_CELL *heap, *heapend;
    long suspended, active, used;
    FL_GOAL **qstart, **qend;
    FL_GOAL **queue;
    long max_goals;
    FL_GOAL *freegoals;
    FL_GOAL *goal_buffer;
    FL_VAL *sstack;
    int ordinal;
    int logging;
    long goals;
    FL_VAL idle;
    unsigned char *port;
    FL_VAL events;
    FL_VAL listening;
#ifdef __linux__
    struct pollfd *pollfds;
#else
    int event_queue;
#endif
    int peers[ 6 ]; /* fwd, bwd, north, east, south, west */
    long delay;  /* microseconds << 8 */
    int state;
    int defer;
    int detached;
    FL_VAL deferred, deferred_tail;
    pthread_t thread;
    FL_HTNODE **globals;
} FL_TCB;

typedef struct {
    FL_VAL name;
    long arity;
    void *addr;
} FL_PDEF;

typedef struct fl_module {
    struct fl_module *next;
    long npdefs;
    FL_VAL name;
    FL_VAL *atoms;
    char *names;
    FL_VAL **ratoms;
    FL_PDEF pdefs[0];
} FL_MODULE;

#define FL_NOT_AN_ATOM      1
#define FL_NOT_A_CELL       2
#define FL_NOT_AN_INT       3
#define FL_NOT_A_LIST       4
#define FL_NOT_A_TUPLE      5
#define FL_NOT_A_VAR        6
#define FL_QUEUE_FULL       7
#define FL_NO_MODULE        8
#define FL_NO_PDEF          9
#define FL_NO_CELLS         10
#define FL_NOT_A_NUMBER     11
#define FL_DIV_BY_ZERO      12
#define FL_CANT_CONVERT     13
#define FL_NOT_A_STRING     14
#define FL_IO_ERROR         15
#define FL_NOT_A_PORT       16
#define FL_PORT_CLOSED      17
#define FL_BAD_INDEX        18
#define FL_CANT_UNIFY       19
#define FL_CANT_FORWARD     20
#define FL_CANT_UNFORWARD   21
#define FL_BAD_PEER         22
#define FL_NOT_A_MODULE     23
#define FL_OUT_OF_MEMORY    24
#define FL_BAD_SIGNAL       25
#define FL_TOO_MANY_GOALS   26
#define FL_NO_GLOBAL        27

#define FL_INPUT            1
#define FL_SIGNAL           2
#define FL_TIMEOUT          3
#define FL_CHILD            4

#define FL_BUSY             0
#define FL_IDLE             1
#define FL_FINISHING        2

#define FL_DETACHED         (-1)

extern char *fl_argv[];
extern FL_TCB fl_tcbs[];
extern FL_VAL fl_nil, fl_true, fl_false, fl_colon;
extern void fl_rt_error(FL_TCB *tcb, FL_VAL x, int error);
extern void fl_rt_error2(FL_TCB *tcb, FL_VAL x, FL_VAL y, int error);
extern void fl_assign(FL_TCB *tcb, FL_VAL x, FL_VAL var);
extern void fl_unify_result(FL_TCB *tcb, FL_VAL val, FL_VAL dest);
extern void fl_release(FL_TCB *tcb, FL_VAL x);
extern void fl_error(FL_TCB *tcb, FL_VAL err);
extern FL_VAL fl_alloc_cell(FL_TCB *tcb, FL_VAL car, FL_VAL cdr);
extern FL_VAL fl_alloc_float(FL_TCB *tcb, double x);
extern void fl_write(FL_TCB *tcb, FILE *fp, FL_VAL x);
extern void fl_fail(FL_TCB *tcb);
extern void fl_logfmt(FL_TCB *tcb, char *fstr, ...);
extern void fl_log_entry(FL_TCB *tcb, long arity);
extern FL_VAL fl_intern(char *np);
extern FL_VAL fl_mkstring(char *str, int len);
extern void fl_init_module(FL_MODULE *mod);
extern void fl_init_goal(FL_PDEF *pdef);
extern void fl_resolve_pdef(FL_TCB *tcb, long arity, FL_VAL mod, FL_VAL name, 
    void *cache, FL_GOAL *ng);
extern FL_VAL fl_unify_rec(FL_TCB *tcb, FL_VAL x, FL_VAL y);
extern void fl_yield(FL_TCB *tcb);
extern void fl_enter(FL_TCB *tcb);
extern void fl_suspend(FL_TCB *tcb);
extern void fl_start(int threads, long heapsz, int maxgoals);
extern void fl__init(void);
extern int fl_ordering(FL_VAL x, FL_VAL y);
extern FL_VAL fl_ordered_below(FL_TCB *ignored, FL_VAL x, FL_VAL y);
extern FL_VAL fl_chase_tail(FL_VAL st);
extern void fl_add_event(FL_TCB *tcb, FL_VAL event);
extern int fl_forward(FL_TCB *tcb, int id, FL_VAL x);
extern int fl_forward_to_port(FL_TCB *tcb, unsigned char *port, FL_VAL x, int dest);
extern FL_VAL fl_unforward(FL_TCB *tcb);
extern void fl_send_to_port(FL_TCB *tcb, FL_VAL port, FL_VAL x);
extern int fl_nthreads(void);
extern int fl_create_event_queue(FL_TCB *tcb);
extern int fl_wait_for_events(FL_TCB *tcb, long ms);
extern int fl_add_event_to_queue(FL_TCB *tcb, int event, FL_VAL data, FL_VAL var,
    FL_VAL eid);
extern void fl_cancel_timer(FL_TCB *tcb, FL_VAL id);
extern FL_MODULE *fl_find_module(FL_TCB *tcb, FL_VAL name);
extern double fl_random(void);
extern void fl_set_random_seed(unsigned char *buf, int n);

#endif
