; FLENG - asm part of x86_64 runtime

    BITS 64
    SECTION .text

    extern fl_rt_error
    extern fl_release
    extern fl_resolve_pdef
    extern fl_true, fl_false, fl_nil
    extern fl_log_entry, fl_writeln
    extern fl_unify_rec
    extern fl_ordering

    ; rdx=error-code, rsi=argument
error:
    mov rdi, rbx
    and rsp, -16
    call fl_rt_error    

    global fl_check_atom
    ; rax=val
fl_check_atom:
    mov rdx, rax
    and rdx, 3
    cmp rdx, 2
    jne .l1
    ret
.l1:
    mov rsi, rax
    mov rdx, 1          ; FL_NOT_AN_ATOM
    jmp error

    global fl_check_cell
    ; rax=val
fl_check_cell:
    test rax, 3
    jnz .l1
    ret
.l1:
    mov rsi, rax
    mov rdx, 2          ; FL_NOT_A_CELL
    jmp error

    global fl_check_integer
    ; rax=val
fl_check_integer:
    test rax, 1         ; FL_INT_BIT
    jz .l1
    ret
.l1:
    mov rsi, rax
    mov rdx, 3          ; FL_NOT_AN_INT
    jmp error

    global fl_check_list
    ; rax=val
fl_check_list:
    test rax, 3
    jnz .l1
    mov rdx, [rax]
    mov rcx, 0xff000000 ; FL_TAG_MASK
    and rdx, rcx
    cmp rdx, 0x01000000 ; FL_LIST_TAG
    jne .l1
    ret
.l1:
    mov rsi, rax
    mov rdx, 4          ; FL_NOT_A_LIST
    jmp error

    global fl_check_module
    ; rax=val
fl_check_module:
    test rax, 3
    jnz .l1
    mov rdx, [rax]
    mov rcx, 0xff000000 ; FL_TAG_MASK
    and rdx, rcx
    cmp rdx, 0x08000000 ; FL_MODULE_TAG
    jne .l1
    ret
.l1:
    mov rsi, rax
    mov rdx, 23          ; FL_NOT_A_MODULE
    jmp error

    global fl_check_tuple
    ; rax=val
fl_check_tuple:
    test rax, 3
    jnz .l1
    mov r9d, [rax]
    test r9d, 0x80000000  ; FL_TUPLE_BIT
    jnz .l1
    ret
.l1:
    mov rsi, rax
    mov rdx, 5          ; FL_NOT_A_TUPLE
    jmp error

    global fl_check_var
    ; rax=val
fl_check_var:
    test rax, 3
    jnz .l1
    mov edx, [rax]
    and edx, 0xff000000 ; FL_TAG_MASK
    cmp edx, 0x04000000 ; FL_VAR_TAG
    je .l1
    cmp edx, 0x07000000 ; FL_REF_TAG
    je .l1
    mov rsi, rax
    mov rdx, 6          ; FL_NOT_A_VAR
    jmp error
.l1:
    ret

    global fl_check_number
    ; rax=val
fl_check_number:
    test rax, 1
    jz .l1
    ret
.l1:
    test rax, 2
    jnz .l2
    mov edx, [rax]
    and edx, 0xff000000 ; FL_TAG_MASK
    cmp edx, 0x02000000 ; FL_FLOAT_TAG
    jne .l2
    ret
.l2:
    mov rsi, rax
    mov rdx, 11         ; FL_NOT_A_NUMBER
    jmp error

    ; rsi=val -> xmm0
float_arg_1:
    test rsi, 1
    jz .l1
    sar rsi, 1
    cvtsi2sd xmm0, rsi
    ret
.l1:
    movsd xmm0, [rsi + 8]
    ret

    ; rdx=val -> xmm1
float_arg_2:
    test rdx, 1
    jz .l1
    sar rdx, 1
    cvtsi2sd xmm1, rdx
    ret
.l1:
    movsd xmm1, [rdx + 8]
    ret

    ; xmm0=val -> rax (float)
float_result:
    mov rax, 1
    mov r10, rax
    call cons
    movsd [rax + 8], xmm0
    mov dword [rax], 0x02000000     ; FL_FLOAT_TAG
    ret

    global fl_add
    ; rsi=num, rdx=num -> rax
fl_add:
    test rsi, 1
    jz .l1
    test rdx, 1
    jz .l1
    lea rax, [rsi - 1]
    add rax, rdx
    ret
.l1:
    call float_arg_1
    call float_arg_2
    addsd xmm0, xmm1
    jmp float_result

    global fl_sub
    ; rsi=num, rdx=num -> rax
fl_sub:
    test rsi, 1
    jz .l1
    test rdx, 1
    jz .l1
    sub rsi, rdx
    lea rax, [rsi + 1]
    ret
.l1:
    call float_arg_1
    call float_arg_2
    subsd xmm0, xmm1
    jmp float_result

    global fl_mul
    ; rsi=num, rdx=num -> rax
fl_mul:
    test rsi, 1
    jz .l1
    test rdx, 1
    jz .l1
    sar rsi, 1
    sar rdx, 1
    imul rsi, rdx
    mov rax, rsi
    shl rax, 1
    or rax, 1
    ret
.l1:
    call float_arg_1
    call float_arg_2
    mulsd xmm0, xmm1
    jmp float_result

    global fl_div
    ; rsi=num, rdx=num -> rax
fl_div:
    cmp rdx, 1
    jne .l0
    mov rsi, 12         ; FL_DIV_BY_ZERO
    jmp error
.l0:
    test rsi, 1
    jz .l1
    test rdx, 1
    jz .l1
    sar rsi, 1
    sar rdx, 1
    mov rax, rsi
    mov rcx, rdx
    cqo
    idiv rcx
    shl rax, 1
    or rax, 1
    ret
.l1:
    call float_arg_1
    call float_arg_2
    divsd xmm0, xmm1
    jmp float_result

    global fl_mod
    ; rsi=num, rdx=num -> rax
fl_mod:
    cmp rdx, 1
    jne .l0
    mov rsi, 12         ; FL_DIV_BY_ZERO
    jmp error
.l0:
    test rsi, 1
    jz .l1
    test rdx, 1
    jz .l1
    sar rsi, 1
    sar rdx, 1
    mov rax, rsi
    mov rcx, rdx
    cqo
    idiv rcx
    shl rdx, 1
    or rdx, 1
    mov rax, rdx
    ret
.l1:
    call float_arg_1
    call float_arg_2
    divsd xmm0, xmm1
    jmp float_result

    global fl_sqrt
    ; rsi=num -> rax
fl_sqrt:
    call float_arg_1
    movsd [rsp - 8], xmm0
    fld qword [rsp - 8]
    fsqrt
    fstp qword [rsp - 8]
    movsd xmm0, [rsp - 8]
    jmp float_result

    global fl_sin
    ; rsi=num -> rax
fl_sin:
    call float_arg_1
    extern sin
    push rax    ; dummy
    call sin
    add rsp, 8
    jmp float_result

    global fl_cos
    ; rsi=num -> rax
fl_cos:
    call float_arg_1
    extern cos
    push rax    ; dummy
    call cos
    add rsp, 8
    jmp float_result

    global fl_truncate
    ; rsi=num -> rax
fl_truncate:
    call float_arg_1
    extern trunc
    push rax    ; dummy
    call trunc
    add rsp, 8
    jmp float_result

    global fl_floor
    ; rsi=num -> rax
fl_floor:
    call float_arg_1
    extern floor
    push rax    ; dummy
    call floor
    add rsp, 8
    jmp float_result

    global fl_round
    ; rsi=num -> rax
fl_round:
    call float_arg_1
    extern round
    push rax    ; dummy
    call round
    add rsp, 8
    jmp float_result

    global fl_ceiling
    ; rsi=num -> rax
fl_ceiling:
    call float_arg_1
    extern ceil
    push rax    ; dummy
    call ceil
    add rsp, 8
    jmp float_result

    global fl_shl
    ; rsi=int, rdx=int -> rax(int)
fl_shl:
    mov rax, rsi
    mov rcx, rdx
    shr rcx, 1
    and rax, ~1
    shl rax, cl
    or rax, 1
    ret

    global fl_shr
    ; rsi=int, rdx=int -> rax(int)
fl_shr:
    mov rax, rsi
    mov rcx, rdx
    shr rcx, 1
    sar rax, cl
    or rax, 1
    ret

    global fl_int
    ; rsi=num -> rax
fl_int:
    test rsi, 1
    jz .l1
    mov rax, rsi
    ret
.l1:
    fld qword [rsi + 8]
    fisttp qword [rsp - 8]
    mov rax, [rsp - 8]
    shl rax, 1
    or rax, 1
    ret

    global fl_float
    ; rsi=num -> rax
fl_float:
    call float_arg_1
    call float_result
    ret

true:
    mov rax, [fl_true]
    ret

false:
    mov rax, [fl_false]
    ret

    global fl_sametype
    ; rsi=val, rdx=val -> rax
fl_sametype:
    mov rax, rsi
    and rax, 3
    mov r9, rdx
    and r9, 3
    cmp rax, r9
    jne false
    test rax, 3
    jnz true
    mov eax, [rsi]
    mov r9d, [rdx]
    test eax, 0x80000000    ; FL_TUPLE_BIT
    jz .l4
    test r9d, 0x80000000    ; FL_TUPLE_BIT
    jnz false
    jmp true
.l4:
    test r9d, 0x80000000    ; FL_TUPLE_BIT
    jnz false
    and eax, 0xff000000     ; FL_TAG_MASK
    and r9d, 0xff000000     ; FL_TAG_MASK
    cmp eax, r9d
    jne false
    jmp true

    global fl_equal
    ; rsi=val, rdx=val -> rax
    ; must preserve P0-P3
fl_equal:
    cmp rsi, rdx
    je true
    test rsi, 1    
    jz .l2
    test rdx, 1
    jz .l3
    jmp false
.l2:
    test rsi, 2
    jnz false
    mov r9d, [rsi]
    and r9d, 0xff000000     ; FL_TAG_MASK
    cmp r9d, 0x02000000     ; FL_FLOAT_TAG
    jne false
.l3:
    test rdx, 2
    jnz false
    test rdx, 1
    jnz .l5
    mov r9d, [rdx]
    and r9d, 0xff000000     ; FL_TAG_MASK
    cmp r9d, 0x02000000     ; FL_FLOAT_TAG
    jne false
.l5:
    call float_arg_1
    call float_arg_2
    ucomisd xmm0, xmm1
    jne false
    jmp true
    
    global fl_less
    ; rsi=val, rdx=val -> rax
    ; must preserve P0-P3
fl_less:
    test rsi, 1
    jz .l1
    test rdx, 1
    jz .l1
    mov rax, [fl_true]
    cmp rsi, rdx
    cmovge rax, [fl_false]
    ret
.l1:
    call float_arg_1
    call float_arg_2
    mov rax, [fl_true]
    ucomisd xmm0, xmm1
    cmovae rax, [fl_false]
    ret

    global fl_below
    ; rsi=val, rdx=val -> rax
    ; must preserve P0-P3
fl_below:
    push rsi
    push rdx
    push rcx
    push r8
    mov rdi, rsi
    mov rsi, rdx
    call fl_ordering
    pop r8
    pop rcx
    pop rdx
    pop rsi
    test rax, 2     ; -1
    jnz true
    jmp false

    global fl_above
    ; rsi=val, rdx=val -> rax
    ; must preserve P0-P3
fl_above:
    push rsi
    push rdx
    push rcx
    push r8
    mov rdi, rsi
    mov rsi, rdx
    call fl_ordering
    pop r8
    pop rcx
    pop rdx
    pop rsi
    cmp rax, 1
    jne false
    jmp true

    global fl_deref
    ; rax=val -> rax
    ; must preserve P0-P3
fl_deref:
    test rax, 3
    jz .l1
.l2:
    ret
.l1:
    mov r9d, [rax]
    and r9d, 0xff000000 ; FL_TAG_MASK
    cmp r9d, 0x04000000 ; FL_VAR_TAG
    jne .l2
    mov r9, [rax + 8]
    cmp rax, r9
    je .l2
    mov rax, r9
    jmp fl_deref

    global fl_unref
    ; r9=val, rbx=tcb
    ; must preserve A + P0-P3
fl_unref:
    test r9, 3
    jz .l1
    ret
.l1:
    mov r10d, dword [r9]
    and r10d, 0xffffff      ; FL_COUNT_MASK
    cmp r10d, 0xffffff      ; FL_COUNT_MASK
    jne .l3
    ret
.l3:
    dec r10d
    jz .l2
    dec dword [r9]
    ret
.l2:
    push rax
    push rsi
    push rdx
    push rcx
    push r8
    mov rdi, rbx
    mov rsi, r9
    call fl_release
    pop r8
    pop rcx
    pop rdx
    pop rsi
    pop rax
    ret

setcarry:
    stc
    ret

clrcarry:
    clc
    ret

    global fl_force
    ; rax=val -> rax, CF is 1 when suspending
    ; rbx=tcb
fl_force:
    test rax, 3
    jnz clrcarry
    mov edx, [rax]
    and edx, 0xff000000 ; FL_TAG_MASK
    cmp edx, 0x07000000 ; FL_REF_TAG
    je .l3
    cmp edx, 0x04000000 ; FL_VAR_TAG
    jne clrcarry
    mov rdx, [rax + 8]
    cmp rax, rdx
    je .l3
    mov rax, rdx
    jmp fl_force
.l3: 
    mov r9, [rbx + 8]   ; tcb->sstackp
    mov [r9], rax
    mov rax, [r12]      ; goal->addr
    mov [r9 + 8], rax
    add r9, 16
    mov [rbx + 8], r9   ; tcb->sstackp
    jmp setcarry

    global fl_islist_t
    ; rax=val, set CF if not a list
fl_islist_t:
    test rax, 3
    jnz setcarry
    mov r9d, [rax]
    and r9d, 0xff000000 ; FL_TAG_MASK
    cmp r9d, 0x01000000 ; FL_LIST_TAG
    jne setcarry
    jmp clrcarry

    global fl_istuple_t
    ; rax=val, set CF if not a tuple
fl_istuple_t:
    test rax, 3
    jnz setcarry
    mov edx, [rax]
    test edx, 0x80000000 ; FL_TUPLE_BIT
    jz setcarry
    jmp clrcarry

    ; rsi=addr, r12=G
trynext:
    mov [r12], rsi      ; goal->addr
    add rsp, 8
    jmp rsi

    global fl_isval
    ; rdx=val, rsi=addr
fl_isval:
    cmp rax, rdx
    jne trynext
    ret

    global fl_isfloat
    ; rax=val, rdx=float, rsi=addr 
fl_isfloat:
    test rax, 3
    jz .l1
    jmp trynext
.l1:
    cmp [rax + 8], rdx
    jne trynext
    ret

    global fl_islist
    ; rax=val, rsi=addr -> rax(car)
    ; push cdr on stack
fl_islist:
    call fl_islist_t
    jc trynext
    mov rdx, [rax + 16]
    xchg [rsp], rdx
    mov rax, [rax + 8]
    jmp rdx

    global fl_isfunctor
    ; rax=val, rsi=addr, rcx=arity, rdx=atom -> rax
fl_isfunctor:
    test rax, 3
    jnz trynext
    mov r8d, [rax]
    test r8d, 0x80000000 ; FL_TUPLE_BIT
    jz trynext
    shr r8d, 24
    and r8d, 0x7f
    cmp r8d, ecx
    jne trynext
    mov rcx, [rax + 8]
    cmp rcx, rdx
    jne trynext
    mov rax, [rax + 16]
    ret

    ; rbx=tcb, r10=car, rax=cdr -> rax
    ; must preserve parameter registers
cons:
    mov r9, [rbx + 6 * 8]   ; tcb->freelist
    cmp r9, 0
    jnz .l1
    mov rdx, 10             ; FL_NO_CELLS
    jmp error
.l1:
    push rax
    mov rax, [r9 + 16]
    mov [rbx + 6 * 8], rax  ; tcb->freelist
    pop rax
    call fl_addref
    mov [r9 + 16], rax
    mov rax, r10
    call fl_addref
    mov [r9 + 8], rax
    inc qword [rbx + 11 * 8]    ; tcb->used
    mov rax, r9
    ret

    global fl_mkfloat
    ; rdx=float64
fl_mkfloat:
    push rdx
    mov rax, 1
    mov r10, 1
    call cons
    pop r9
    mov dword [rax], 0x02000000     ; FL_FLOAT_TAG
    mov [rax + 8], r9
    ret
    
    global fl_mklist
    ; rax=cdr, [rsp+8]=car -> rax
fl_mklist:
    mov r10, [rsp + 8]
    call cons
    mov r9, 0x01000000      ; FL_LIST_TAG
    mov [rax], r9
    pop r10
    add rsp, 8
    jmp r10

    global fl_mktuple
    ; rax=args, r10=atom, r9=arity -> rax
fl_mktuple:
    push r9
    call cons
    pop r9
    shl r9, 24
    or r9d, 0x80000000 ; FL_TUPLE_BIT
    mov [rax], r9d
    ret

    global fl_mkvar
    ; -> rax
fl_mkvar:
    mov r10, 1
    mov rax, [fl_nil]
    call cons
    mov [rax + 8], rax
    mov dword [rax], 0x04000000   ; FL_VAR_TAG
    ret

    ; r13=goal, rbx=tcb
enqueue:
    mov rax, [rbx + 13 * 8]     ; tcb->qend
    mov [rax], r13
    add rax, 8
    mov rcx, [rbx + 12 * 8]     ; tcb->qstart
    cmp rax, rcx
    jne .l1
    mov rdx, 7                  ; FL_QUEUE_FULL
    jmp error
.l1:
    mov rcx, [rbx + 15 * 8]     ; tcb->max_goals
    shl rcx, 3
    add rcx, [rbx + 14 * 8]     ; tcb->queue
    cmp rax, rcx
    jb .l2
    mov rax, [rbx + 14 * 8]     ; tcb->queue
.l2:
    mov [rbx + 13 * 8], rax     ; tcb->qend
    inc qword [rbx + 10 * 8]    ; tcb->active
    ret

    global fl_opengoal
    ; -> r13
fl_opengoal:
    mov r13, [rbx + 16 * 8]     ; tcb->freegoals
    cmp r13, 0
    jne .l1
    mov rdx, 26                 ; FL_TOO_MANY_GOALS
    jmp error
.l1:
    mov r9, [r13]               ; goal->addr
    mov [rbx + 16 * 8], r9      ; tcb->freegoals
    inc qword [rbx + 20 * 8]    ; tcb->goals
    jmp enqueue

    global fl_dropgoal
    ; rbx=tcb
fl_dropgoal:
    mov r9, [rbx]               ; tcb->goal
    mov r10, [rbx + 16 * 8]     ; tcb->freegoals
    mov [r9], r10               ; goal->addr
    mov [rbx + 16 * 8], r9      ; tcb->freegoals
    dec qword [rbx + 20 * 8]    ; tcb->goals
    ret

    global fl_addref
    ; rax=val
fl_addref:
    test rax, 3
    jz .l1
    ret
.l1:
    push rax
    mov eax, dword [rax]
    and eax, 0xffffff      ; FL_COUNT_MASK
    cmp eax, 0xffffff      ; FL_COUNT_MASK
    je .l2
    pop rax
    inc dword [rax]
    ret
.l2:
    pop rax
    ret

    global fl_puttailarg
    ; r13=N, rax=list
fl_puttailarg:
    mov rdx, [rax + 16]
    test rdx, 3
    jz .l1
    mov rax, [rax + 8]
.l1:
    call fl_addref
    mov [r13 + 4 * 8], rax
    ret

    global fl_resolve
    ; r8=cache, rsi=arity, rdx=module, rcx=name
    ; r13=N
fl_resolve:
    mov r9, [r8]
    cmp r9, 0
    je .l1
    mov [r13], r9       ; goal->addr
    ret
.l1:
    mov rdi, rbx
    mov r9, r13
    jmp fl_resolve_pdef

    global fl_replace
    ; rax=val, rdx=argaddr
fl_replace:
    cmp rax, [rdx]
    jne .l1
    ret
.l1:
    mov r9, [rdx]
    call fl_addref
    mov [rdx], rax
    jmp fl_unref
    
    global fl_tail_goal
    ; rdx=addr, rbx=tcb -> r13(N)
fl_tail_goal:
    mov rcx, [rbx + 16] ; tcb->tailcalls
    mov rsi, [rbx + 24] ; tcb->timeslice
    cmp rcx, rsi
    jl fl_invoke_goal
    push rdx
    call fl_opengoal
    pop rdx
    mov [r13], rdx      ; goal->addr
    ret

    global fl_invoke_goal
    ; rdx=addr, rbx=tcb -> r13(N)
fl_invoke_goal:
    mov r13, r12
    mov [r13], rdx      ; goal->addr
    ret

    global fl_tail_jump
    ; r13=r12=N, rbx=tcb, rsp=frame
fl_tail_jump:
    mov rcx, [rbx + 16] ; tcb->tailcalls
    mov rsi, [rbx + 24] ; tcb->timeslice
    cmp rcx, rsi
    jl .l1
    xor rax, rax
    mov [rbx + 16], rax ; tcb->tailcalls
    ret
.l1:
    inc rcx
    mov [rbx + 16], rcx ; tcb->tailcalls
    ; fall through ...

    global fl_invoke_jump
    ; r13=r12=N, rbx=tcb, rsp=frame
fl_invoke_jump:
    mov rsp, rbp
    mov r12, r13
    mov [rbx], r13      ; tcb->goal
    mov rax, [r13]      ; goal->addr
    jmp rax

    global fl_test_integer
    ; rsi=val -> rax (bool)
fl_test_integer:
    test rsi, 1
    jnz true
    jmp false

    global fl_test_number
    ; rsi=val -> rax (bool)
fl_test_number:
    test rsi, 1
    jnz true
    mov r9d, 0x02000000     ; FL_FLOAT_TAG
    jmp fl_test_tag

    global fl_test_atom
    ; rsi=val -> rax (bool)
fl_test_atom:
    mov rax, rsi
    and rax, 3
    cmp rax, 2
    jne false
    jmp true

    global fl_test_tag
    ; rsi=val, r9d=tag -> rax (bool)
fl_test_tag:
    test rsi, 3
    jnz false
    mov eax, [rsi]
    and eax, 0xff000000 ; FL_TAG_MASK
    cmp eax, r9d
    jne false
    jmp true

    global fl_test_var
    ; rsi=val -> rax (bool)
fl_test_var:
    test rsi, 3
    jnz false
    mov eax, [rsi]
    and eax, 0xff000000 ; FL_TAG_MASK
    cmp eax, 0x07000000 ; FL_REF_TAG
    je true
    cmp eax, 0x04000000 ; FL_VAR_TAG
    jne false
    mov rax, [rsi + 8]
    cmp rax, rsi
    jne false
    jmp true

    global fl_test_tuple
    ; rsi=val -> rax (bool)
fl_test_tuple:
    test rsi, 3
    jnz false
    mov eax, [rsi]
    test eax, 0x80000000 ; FL_TUPLE_BIT
    jz false
    jmp true

    global fl_enter
    ; rdi=tcb
fl_enter:
    mov rsp, [rdi + 5 * 8]  ; tcb->stackbase
    mov rbp, rsp
    mov r12, [rdi]      ; tcb->goal
    mov rax, [r12]      ; goal->addr
    mov rbx, rdi
    jmp rax

    global fl_atomic
    ; rsi=val -> rax (bool)
fl_atomic:
    test rsi, 3         ; FL_BITS_MASK
    jnz true
    mov eax, [rsi]
    and eax, 0xff000000     ; FL_TAG_MASK
    cmp eax, 0x02000000     ; FL_FLOAT_TAG
    je true
    jmp false

    global fl_entry
    ; rbx=tcb, rsi=arity
fl_entry:
    mov eax, [rbx + 19 * 8]     ; tcb->logging
    test eax, 1
    jnz .l1
    ret
.l1:
    mov rdi, rbx
    jmp fl_log_entry
    ret

    global fl_unify
    ; rsi=val, rdx=val -> rax (bool)
    ; must preserve P0-P3
fl_unify:
    push rsi
    push rdx
    push rcx
    push r8
    mov rdi, rbx
    sub rsp, 8              ; (un)align
    call fl_unify_rec
    add rsp, 8
    pop r8
    pop rcx
    pop rdx
    pop rsi
    ret
